/*jslint vars: true, plusplus: true, devel: true, nomen: true, regexp: true, indent: 4, maxerr: 50 */
/*global $, window, location, CSInterface, SystemPath, themeManager*/

(function () {
    'use strict';

	var debug = false;
	var debug_stack = new Array();
	var debug_buffer = "";
	var release_make_persistent = true; // true for release, false for debugging

    var csInterface = new CSInterface();
	var mData = {};
    
    // Opens the chrome developer tools in host app
    function showDevTools() {
        window.__adobe_cep__.showDevTools();
    }
    
	function ReloadJs(scriptName) {
		var docHeadObj = document.getElementsByTagName("head")[0];
		var dynamicScript = document.createElement("script");
		dynamicScript.type = "text/javascript";
		dynamicScript.src = scriptName;
		docHeadObj.appendChild(newScript);
	}

    // Reloads extension panel
    function reloadPanel() {
//		loadJSXFile("jsx/hostscript.jsx");
        location.reload();
		ReloadJs("js/main.js");
    }
    
    // Loads / executes a jsx file
    function loadJSXFile(pPath) {
        var scriptPath = csInterface.getSystemPath(SystemPath.EXTENSION) + pPath;
        csInterface.evalScript('$._ext.evalFile("' + scriptPath + '")');
    }
    
    // Loads / executes all jsx files in the given folder
    function loadJSXFiles(pFolderPath) {
        var extensionRoot = csInterface.getSystemPath(SystemPath.EXTENSION) + pFolderPath;
        csInterface.evalScript('$._ext.evalFiles("' + extensionRoot + '")');
    }
    
	function Alert(msg)
	{
		csInterface.evalScript('$._ext.Alert("'+msg+'")');
	}

	function DBG(msg) {
		if (debug) {
			//alert('DBG:'+msg);
			if (msg.length==1) {
				// Use a buffer for short messages, because if the file monitor outputs a '.' every half second, then it is difficult to select the debug textarea and copy its contents.
				debug_buffer += msg;
			} else {
				// clear the buffer
				if (debug_buffer != "") {
					$("#dbg_out").append(debug_buffer + "\n");
					debug_buffer = "";
				}

				var s = "";
				for (var i=0; i<debug_stack.length; i++) {
					s += debug_stack[i] + ":";
				}
				$("#dbg_out").append(s+' '+msg+"\n");
			}
		}
	};
	function DBG_PUSH(name) {
		debug_stack.push(name);
		DBG("START");
	}
	function DBG_POP() {
		DBG("END");
		debug_stack.pop();
	}

	function file_exists(path) {
		var stat = window.cep.fs.stat(path);
		//DBG("file_exists("+path+") stat:"+JSON.stringify(stat));
		//DBG("file date:"+result.data.mtime);
		return stat.err == 0;
	}
	function file_delete(path)
	{
		DBG_PUSH("file_delete()");
		if (!file_exists(path)) { DBG("the file to delete does not exist."); return; }
		var result = window.cep.fs.deleteFile(path);
		if (result.err != 0) { DBG("unable to delete the file '"+path+"'"); }
		if (!file_exists(path)) { DBG("delete seems to be successful: the file does not exist after the delete."); }
		DBG_POP();
	}
	function file_rename(from, to)
	{
		DBG_PUSH("file_rename()");
		if (!file_exists(from)) { DBG("the file to rename does not exist."); return; }
		DBG("from:" + from);
		DBG("to  :" + to  );
		if (file_exists(to)) {
			DBG("WARNING: target file already exists, thus removing it.");
			file_delete(to);
		}
		var result = window.cep.fs.rename(from, to);
		if (result.err != 0) { DBG("unable to rename the file '"+from+"' to '"+to+"'"); }
		if (!file_exists(from)) { DBG("rename seems to be successful: the FROM file does not exist after the rename."); }
		DBG_POP();
	}

	function RefreshPath()
	{
		console.log("RefreshPath()");
		DBG_PUSH("RefreshPath()");
		DBG("RefreshPath() BEGIN 1");
		// Path should be refreshed every time because of switching between currently running versions of Escape Motions apps
		DBG("mData:"+JSON.stringify(mData));

		var paths_in = new Array(
			{'app': 'rebelle', 'macAppId': 'com.escapemotions.rebelle'         , 'appName': 'Rebelle 7'          , 'f_em2ps': 'em2photoshop'   , 'f_ps2em': 'photoshop2em'   , 'f_alive': 'em_alive'   },
			{'app': 'rebelle', 'macAppId': 'com.escapemotions.rebelle'         , 'appName': 'Rebelle 6'          , 'f_em2ps': 'em2photoshop'   , 'f_ps2em': 'photoshop2em'   , 'f_alive': 'em_alive'   },
			{'app': 'rebelle', 'macAppId': 'com.escapemotions.rebelle'         , 'appName': 'Rebelle 5'          , 'f_em2ps': 'em2photoshop'   , 'f_ps2em': 'photoshop2em'   , 'f_alive': 'em_alive'   },
			{'app': 'rebelle', 'macAppId': 'com.escapemotions.rebelle'         , 'appName': 'Rebelle 5 Beta 8'   , 'f_em2ps': 'em2photoshop'   , 'f_ps2em': 'photoshop2em'   , 'f_alive': 'em_alive'   },
			{'app': 'rebelle', 'macAppId': 'com.escapemotions.rebelle'         , 'appName': 'Rebelle 5 Beta 7'   , 'f_em2ps': 'em2photoshop'   , 'f_ps2em': 'photoshop2em'   , 'f_alive': 'em_alive'   },
			{'app': 'rebelle', 'macAppId': 'com.escapemotions.rebelle'         , 'appName': 'Rebelle 5 Beta 6'   , 'f_em2ps': 'em2photoshop'   , 'f_ps2em': 'photoshop2em'   , 'f_alive': 'em_alive'   },
			{'app': 'rebelle', 'macAppId': 'com.escapemotions.rebelle'         , 'appName': 'Rebelle 5 Beta 5'   , 'f_em2ps': 'em2photoshop'   , 'f_ps2em': 'photoshop2em'   , 'f_alive': 'em_alive'   },
			{'app': 'rebelle', 'macAppId': 'com.escapemotions.rebelle'         , 'appName': 'Rebelle 5 Beta 4'   , 'f_em2ps': 'em2photoshop'   , 'f_ps2em': 'photoshop2em'   , 'f_alive': 'em_alive'   },
			{'app': 'rebelle', 'macAppId': 'com.escapemotions.rebelle'         , 'appName': 'Rebelle 5 Beta 3'   , 'f_em2ps': 'em2photoshop'   , 'f_ps2em': 'photoshop2em'   , 'f_alive': 'em_alive'   },
			{'app': 'rebelle', 'macAppId': 'com.escapemotions.rebelle'         , 'appName': 'Rebelle 5 Beta 2'   , 'f_em2ps': 'em2photoshop'   , 'f_ps2em': 'photoshop2em'   , 'f_alive': 'em_alive'   },
			{'app': 'rebelle', 'macAppId': 'com.escapemotions.rebelle'         , 'appName': 'Rebelle 5 Beta'     , 'f_em2ps': 'em2photoshop'   , 'f_ps2em': 'photoshop2em'   , 'f_alive': 'em_alive'   },
			{'app': 'rebelle', 'macAppId': 'com.escapemotions.rebelle'         , 'appName': 'Rebelle 4'          , 'f_em2ps': 'em2photoshop'   , 'f_ps2em': 'photoshop2em'   , 'f_alive': 'em_alive'   },
			{'app': 'flame'  , 'macAppId': 'com.escapemotions.flamepainter3pro', 'appName': 'Flame Painter 4'    , 'f_em2ps': 'flame2photoshop', 'f_ps2em': 'photoshop2flame', 'f_alive': 'flame_alive'},
			{'app': 'flame'  , 'macAppId': 'com.escapemotions.flamepainter3pro', 'appName': 'Flame Painter 3 Pro', 'f_em2ps': 'flame2photoshop', 'f_ps2em': 'photoshop2flame', 'f_alive': 'flame_alive'});
		DBG("paths_in:"+JSON.stringify(paths_in));

		mData.paths = new Array();
		var f_add_path = function(rec, path){
			DBG('f_add_path() path:'+path);
			if (file_exists(path)) {
				DBG("f_add_path() path exists -> adding");
				var rec_copy = Object.assign({}, rec);
				rec_copy['path'] = path;
				rec_copy['alive'] = file_exists(path + '/' + rec['f_alive'] + '.dat');
				DBG("alive:"+rec_copy['alive']);
				mData.paths.push(rec_copy);
			}
		};
		if (mData.isWindows) {
			DBG("_3");
			for (var i=0; i<paths_in.length; i++) {
				DBG("i:" + i);
				f_add_path(paths_in[i], mData.homedir+"/AppData/Local/Escape Motions/"+paths_in[i]['appName']+"/Data/Photoshop_plugin");
				f_add_path(paths_in[i], mData.homedir+"/Local Settings/Application Data/Escape Motions/"+paths_in[i]['appName']+"/Data/Photoshop_plugin");
			}
			//var f = new Folder(mData.homedir+"/AppData/Local/Escape Motions");
			//f.getFiles();
		} else {
			DBG("_4");
			for (var i=0; i<paths_in.length; i++) {
				f_add_path(paths_in[i], mData.homedir+"/Library/Application Support/Escape Motions/"+paths_in[i]['appName']+"/Data/Photoshop_plugin");
				f_add_path(paths_in[i], mData.homedir+"/Library/Containers/"+paths_in[i]['macAppId']+"/Data/Library/Application Support/Escape Motions/"+paths_in[i]['appName']+"/Data/Photoshop_plugin");
			}
		}
		DBG("paths:"+JSON.stringify(mData.paths, null, 2));

		// Remove non-existing paths
		// for (var i=0; i<mData.paths.length; i++) {
		// 	var path = mData.paths[i]['path'];
		// 	DBG("i:"+i+" exists:"+file_exists(path)+" path:"+path);
		// 	if (!file_exists(path)) {
		// 		DBG("path does not exist, thus discard it:"+path);
		// 		// DBG("paths before:"+JSON.stringify(mData.paths, null, 2));
		// 		mData.paths.splice(i, 1);
		// 		i = i-1;
		// 		// DBG("paths after:"+JSON.stringify(mData.paths, null, 2));
		// 	}
		// }
		// DBG("only existing paths:" + JSON.stringify(mData.paths, null, 2));

		// If there is a path containing 'em_alive.dat', choose that one
		// for (var i=0; i<mData.paths.length; i++) {
		// 	var path = mData.paths[i]['path'] + '/em_alive.dat';
		// 	DBG("exists:"+file_exists(path)+" path:"+path);
		// 	if (file_exists(path)) {
		// 		DBG("em_alive.dat found in path, thus forcing the path:"+mData.paths[i]['path']);
		// 		DBG("1");
		// 		mData.paths = new Array(mData.paths[i]);
		// 		DBG("2");
		// 		break;
		// 	}
		// }

		mData.paths.sort((a, b) => {
			if (a.alive && !b.alive) {
				return -1;
			} else if (!a.alive && b.alive) {
				return 1;
			} else {
				return 0;
			}
		});
		DBG("sorted paths:"+JSON.stringify(mData.paths, null, 2));
		
		{
			var apps_set = {};
			for (var i=0; i<mData.paths.length; i++) {
				var rec = mData.paths[i];
				apps_set[rec['app']] = 1;
			}
			DBG("apps_set:"+JSON.stringify(apps_set));
			$('#btn_export_rebelle').prop("disabled", apps_set["rebelle"] === undefined);
			$('#btn_export_flame'  ).prop("disabled", apps_set["flame"  ] === undefined);
			mData.validApps = Object.keys(apps_set);
			DBG("mData.validApps:"+JSON.stringify(mData.validApps));
		}

		DBG("_4");
		// If em_alive.dat was not found, return the first existing path:
		if (mData.paths.length == 0) {
			mData.paths = new Array("");
			DBG("WARNING: no path found");
		}
		DBG("_5");
		DBG("calling PrintData(1)");
		csInterface.evalScript('$._ext.PrintData()', function(result){
			DBG("PrintData(1): "+result);
		});
		DBG("RefreshPath() END");
		DBG_POP();
	}

	function IsEmAppAlive(rec)
	{
		DBG_PUSH("IsEmAppAlive()");
		var res = rec['path'] != "" && file_exists(rec['path'] + '/' + rec['f_alive'] + '.dat');
		DBG("res:"+res+" path:"+rec['path']);
		DBG_POP();
		return res;
	}
	
	var mListenToStdOutFinished = new Array();
	function ListenToStdOut(pid, callback)  
	{
		DBG_PUSH("ListenToStdOut()");
		window.cep.process.stdout(pid, function(output) {
			//DBG(output);
			mListenToStdOutFinished[pid] = callback(output);
		});
		var result = window.cep.process.isRunning(pid);
		if (result.data == true && !mListenToStdOutFinished[pid])  {
  			setTimeout(function() {ListenToStdOut(pid, callback)}, 500);
		} else {
			mListenToStdOutFinished[pid] = false;
		}
		DBG_POP();
	}
	
	var mListenToStdErrFinished = new Array();
	function ListenToStdErr(pid, callback)
	{
		DBG_PUSH("ListenToStdErr()");
		var finished = false;
		window.cep.process.stderr(pid, function(output) {
			//DBG(output);
			mListenToStdErrFinished[pid] = callback(output);
		});
		var result = window.cep.process.isRunning(pid);
		if (result.data == true && !mListenToStdErrFinished[pid])  {
  			setTimeout(function() {ListenToStdErr(pid, callback)}, 500);
		} else {
			mListenToStdErrFinished[pid] = false;
		}
		DBG_POP();
	}

	function RequireEmAppRunning(rec, callback_onIsRunning)
	{
		DBG_PUSH("RequireEmAppRunning()");
		DBG("RequireEmAppRunning()");
		if (IsEmAppAlive(rec)) {
			DBG("RequireEmAppRunning() IsEmAppAlive()==true");
			//alert("Escape Motions application seems not to be running. Please run it first.");
			callback_onIsRunning();
			DBG_POP();
			return;
		}

		var app_location_filename = rec['path'] + "/../app_location.dat";
		DBG("app_location_filename: "+app_location_filename);
		if (!file_exists(app_location_filename)) {
			Alert("Escape Motions application seems not to be running. Please run it first.");
			DBG_POP();
			return;
		}
		
		var app_location = window.cep.fs.readFile(app_location_filename).data;//.trim();
		DBG("app_location: "+app_location);
		DBG("3");
		if (!file_exists(app_location)) {
			Alert("Escape Motions application seems not to be running. Please run it first.");
			DBG_POP();
			return;
		}

		var createProcessResult = window.cep.process.createProcess(app_location);
		DBG("createProcessResult: "+JSON.stringify(createProcessResult));
		if (createProcessResult.err != 0) {
			Alert("Unable to run an Escape Motions application");
			DBG_POP();
			return;
		}
		var pid = createProcessResult.data;
		var onNewData = function(data) {
			//DBG("callback: "+data);
			var lines = data.split("\n");
			DBG("lines: "+JSON.stringify(lines));
			for (var i=0; i<lines.length; i++) {
				//var regexp:RegExp = /SetupPainterCore\(\): Timer t38/; // regexp for os detection
				//if (regexp.test(s)) {
				//DBG("out: "+s.trim());
				if (lines[i].trim() == "PainterArtworkInitialized()") {
					DBG("Application is running and ready to use 1");
					RefreshPath();
					DBG("Application is running and ready to use 2");
					DBG("RequireEmAppRunning() callback()");
					callback_onIsRunning();
					callback_onIsRunning = function() {} // run the callback only once
					return true;
				}
			}
			return false;
		};
		ListenToStdOut(pid, onNewData);
		ListenToStdErr(pid, onNewData);
		DBG_POP();
	}

	function initJsxExtension() {
		DBG_PUSH("initJsxExtension()");
		DBG("SystemPath.MY_DOCUMENTS:"+csInterface.getSystemPath(SystemPath.MY_DOCUMENTS)+" exists:"+file_exists(csInterface.getSystemPath(SystemPath.MY_DOCUMENTS)));
		DBG("SystemPath.USER_DATA:"+csInterface.getSystemPath(SystemPath.USER_DATA)+" exists:"+file_exists(csInterface.getSystemPath(SystemPath.USER_DATA)));
		DBG("SystemPath.COMMON_FILES:"+csInterface.getSystemPath(SystemPath.COMMON_FILES)+" exists:"+file_exists(csInterface.getSystemPath(SystemPath.COMMON_FILES)));
		var homedir = csInterface.getSystemPath(SystemPath.USER_DATA)
					  .replace(/[\/\\]$/, "")
					  .replace(/[\/\\][^\/\\]*$/, "")  // remove last folder from documentsdir
					  .replace(/[\/\\][^\/\\]*$/, ""); // remove last folder from documentsdir
		DBG("initJsxExtension() data before:"+JSON.stringify(mData));
		mData = {
			debug: debug,
			extPath: csInterface.getSystemPath(SystemPath.EXTENSION),
			isWindows: csInterface.getOSInformation().match(/Windows/i)!=null,
			homedir: homedir
		};

		RefreshPath();

		DBG("initJsxExtension()");
		csInterface.evalScript('$._ext.init('+JSON.stringify(mData)+')', function(result){
			DBG("_ext.init() result: "+result);
		});
		DBG("calling PrintData(2)");
		csInterface.evalScript('$._ext.PrintData()', function(result){
			DBG("PrintData(2): "+result);
		});

		// csInterface.evalScript('$._ext.JsonArgTest("aaa")', function(result){
		csInterface.evalScript('$._ext.JsonArgTest('+JSON.stringify(mData)+')', function(result){
			DBG("JsonArgTest1 result: "+result);
		});

		// csInterface.evalScript('$._ext.JsonArgTest("'+JSON.stringify(mData.paths[0])+'")', function(result){
		// 	DBG("JsonArgTest2 result: "+result);
		// });


		DBG_POP();
	}

	function MonitorFileChange(callback_onChanged)
	{
		DBG('.');
		//DBG_PUSH("MonitorFileChange()");
		//mtimeOld = mtimeOld || window.cep.fs.stat(filepath).data.mtime; // set default value
		//DBG("file_exists:"+file_exists(filepath));
		//DBG("mtimeOld:"+mtimeOld);
		//DBG("typeof:"+(typeof mtimeOld));
		//DBG("typeof:"+(typeof mtimeOld !== 'undefined'));
		for (var i=0; i<mData.paths.length; i++) {
			var rec = mData.paths[i];
			var filepath = rec['path'] + '/' + rec['f_em2ps'] + '.dat';

			rec['mtimeOld'] = typeof rec['mtimeOld'] !== 'undefined' ? rec['mtimeOld'] : (file_exists(filepath) ? window.cep.fs.stat(filepath).data.mtime : "file_not_found");
			//DBG("MonitorFileChange()");
			var mtimeNew = file_exists(filepath) ? window.cep.fs.stat(filepath).data.mtime : "file_not_found";
			//DBG("mtimeOld:"+mtimeOld+" mtimeNew:"+mtimeNew+" String(mtimeNew):"+String(mtimeNew));
			if (String(rec['mtimeOld']) != String(mtimeNew) && String(mtimeNew) != "" && String(mtimeNew) != "undefined" && String(mtimeNew) != "file_not_found") {
				DBG("call callback_onChanged()");
				callback_onChanged(rec);
			}
		}
		setTimeout(function(){ MonitorFileChange(callback_onChanged); }, 500);
		//DBG_POP();
	}

	function MakePersistent()
	{
		DBG_PUSH("MakePersistent()");
		var event = new CSEvent("com.adobe.PhotoshopPersistent", "APPLICATION"); // PhotoshopUnPersistent
		event.extensionId = "com.escapemotions.psconnect";
		csInterface.dispatchEvent(event);
		DBG_POP();
	}

	function updateThemeWithAppSkinInfo(appSkinInfo) {
		DBG_PUSH("updateThemeWithAppSkinInfo()");
		var bg = appSkinInfo.panelBackgroundColor.color;
		var grayscale = (bg['red'] + bg['green'] + bg['blue']) / 3;
		if (grayscale < 128) {
			$('body').removeClass('light').addClass('dark');
		} else {
			$('body').removeClass('dark').addClass('light');
		}
		DBG_POP();
	}

	function onAppThemeColorChanged(event) {
		DBG_PUSH("onAppThemeColorChanged()");
		var skinInfo = JSON.parse(window.__adobe_cep__.getHostEnvironment()).appSkinInfo;
		updateThemeWithAppSkinInfo(skinInfo);
		DBG_POP();
	}

	function ExportToApp(appId) {

		csInterface.evalScript('$._ext.PrintData()', function(result){
			DBG("PrintData(3): "+result);
		});
		RefreshPath();

		var rec;
		for (var i=0; i<mData.paths.length; i++) {
			var rec_i = mData.paths[i];
			if (rec_i['app'] == appId) {
				rec = rec_i;
				break;
			}
		}
		if (rec === 'undefined') {
			DBG("no record found");
			return;
		}
		DBG("rec:"+JSON.stringify(rec, null, 2));

		csInterface.evalScript('$._ext.PrintData()', function(result){
			DBG("PrintData(4): "+result);
		});

		{
			DBG("test1");
			var a = "aaa";
			DBG("test1a");
			var path = rec['path'] + "/photoshop_test1.txt";
			DBG("test1 path:"+path);
			result = window.cep.fs.writeFile(path, "This is a test1.");
			DBG("test1 result:"+result.err);
			if (0 == result.err) {
				DBG("test1 success");
				var result = window.cep.fs.readFile(path);
				DBG("test1 read result:"+result.err);
				if (0 == result.err) {
					DBG("test1 read data:"+result.data); //result.data is file content
				} else {
					DBG("test1 read failed");
				}
			} else {
				DBG("test1 failed");
			}
		}

		csInterface.evalScript('$._ext.hasAnyDocument()', function(result){
			//alert("result:"+result);
			if (result == "false") {
				Alert("To export a layer to an Escape Motions application, please create or open a document first.");
			} else {
				RequireEmAppRunning(rec, function(){
					DBG("_ext.exportToEmApp()");
					csInterface.evalScript('$._ext.exportToEmApp(' + JSON.stringify(rec) + ')', function(result){
						//DBG("result:"+JSON.stringify(result));
						//DBG("result:"+result);
						DBG("--------------------------------------------------");
						var result_obj = JSON.parse(result);
						DBG("\n"+result_obj.out);
					});
				});
			}
		});
	}

    function init()
	{
		DBG_PUSH("init()");
		DBG("init()");
        themeManager.init();

		updateThemeWithAppSkinInfo(csInterface.hostEnvironment.appSkinInfo);
		csInterface.addEventListener(CSInterface.THEME_COLOR_CHANGED_EVENT, onAppThemeColorChanged);

		if (debug) {
			$("#content").append("<br/><textarea id='dbg_out'></textarea>");
		}
		if (release_make_persistent) {
			MakePersistent();
		}

		initJsxExtension();

		RefreshPath();

		DBG("MonitorFileChange()");
		MonitorFileChange(function(rec) {
			DBG_PUSH("MonitorFileChange callback()");
			DBG("callback_onChanged()");
			// Rename files to let EM app know that the Photoshop plugin is active
			var path_em2ps = rec['path'] + "/" + rec['f_em2ps'];
			file_rename(path_em2ps + ".dat", path_em2ps + "_processing.dat");
			file_rename(path_em2ps + ".png", path_em2ps + "_processing.png");
			csInterface.evalScript('$._ext.importFromEmApp(' + JSON.stringify(rec) + ')', function(result){
				//DBG("result:"+JSON.stringify(result));
				//DBG("result:"+result);
				DBG("--------------------------------------------------");
				var result_obj = JSON.parse(result);
				DBG("\n"+result_obj.out);
				DBG("delete file:" + path_em2ps + "_processing.dat");
				DBG("delete file:" + path_em2ps + "_processing.png");
				file_delete(path_em2ps + "_processing.dat");
				file_delete(path_em2ps + "_processing.png");
			});
			DBG_POP();
		});

		if (debug) {
			$('ul.toolbar_buttons').append('<li><button id="btn_reload" class="hostBgd hostFontSize hostFontColor hostBorder hostFontFamily">Reload</button></li>');
			$("#btn_reload").click(function () {
				DBG("btn_reload clicked");
				reloadPanel();
			});
		}

        $("#btn_export_rebelle").click(function () {
			DBG("click rebelle");
			ExportToApp("rebelle");
		});

        $("#btn_export_flame").click(function () {
			DBG("click flame");
			ExportToApp("flame");
		});

		DBG_POP();
    }
        
    init();

}());
    
