// Documentation:
// http://www.adobe.com/devnet/photoshop/scripting.html
// http://wwwimages.adobe.com/content/dam/Adobe/en/devnet/photoshop/pdfs/photoshop-cc-javascript-ref-2015.pdf
// File, Folder: http://wiki.ecmascript.org/lib/exe/fetch.php?id=proposals%3Afilesystem&cache=cache&media=proposals:adobefilefolder.pdf

/*jslint vars: true, plusplus: true, devel: true, nomen: true, regexp: true, indent: 4, maxerr: 50 */
/*global $, Folder*/

// debug mode: 0-2 (0:disable, 1:break on error, 2:break at beginning)
$.level = 2;
debugger; // launch debugger on next line

//alert("hostscript init");

$._ext = (function () {
    'use strict';
	//var JSON = require('jsx/JSON');
    
    var ext = {};

	ext.data = {};

	if(typeof String.prototype.trim !== 'function') {
		String.prototype.trim = function() {
			return this.replace(/^\s+|\s+$/g, ''); 
		}
	}

	ext.Alert = function (msg)
	{
		alert(msg);
	}

	ext.PrintData = function ()
	{
		//alert("PrintData(): "+JSON.stringify(ext.data));
//		return JSON.stringify(ext.data);
		var out = "";
		//out += "msg:"+msg+"\n";
		out += "homedir:"+ext.data.homedir+"\n";
		out += "extPath:"+ext.data.extPath+"\n";
		out += "\n";
		out += "\n";
		return out;
	}

/*
	ext.DbgToFile = function (msg)
	{
		f = new File(ext.data.emPluginPath + "/photoshop_html5.log");
		f.encoding = "UTF8"; 
		f.open("a", "TEXT");
		f.write(msg + "\n");
		f.close();
	}
*/

	ext.JsonArgTest = function (obj)
	{
		// var out = "ok\n";
		var out = "JsonArgTest obj:"+obj;
		out += "\n";
		out += "JsonArgTest obj.stringify:"+JSON.stringify(obj);
		out += "\n";
		return out;
	}

	ext.hasAnyDocument = function ()
	{
		return app.documents.length > 0;
	}

	//--------------------------------------------------
	// exportToEmApp()
	//--------------------------------------------------
	var exportToEmApp_var_out = "";
	ext.exportToEmApp = function (rec)
	{
		if (app.documents.length == 0)
		{
			alert("To export a layer to an Escape Motions application, please create or open a document first.");
			return;
		}
		
		app.activeDocument.suspendHistory('Export to an Escape Motions application','ext.exportToEmApp_impl('+JSON.stringify(rec)+');');
		return JSON.stringify({out:exportToEmApp_var_out});
	}
	ext.exportToEmApp_impl = function (rec) {
		var out = "";
		function DBG(msg) {
			if (ext.data.debug) {
				out += 'exportToEmApp(): ' + msg + '\n';
			}
		}
		function DBGA(msg) {
			if (ext.data.debug) {
				if (0) { alert('exportToEmApp(): ' + msg); }
				else   { DBG(msg); }
			}
		}
	
		DBGA("BEGIN");
	
		DBG("rec:"+rec);
		DBG("rec.stringify:"+JSON.stringify(rec, null, 2));
	
		var doc = app.activeDocument;
		var activeLayer = doc.activeLayer;
		DBG("activeLayer:"+activeLayer.name);
		
		// Allow either ArtLayer or LayerSet
		if (activeLayer.typename !== "ArtLayer" && activeLayer.typename !== "LayerSet") {
			alert("Active layer must be a regular layer or layer group ...return");
			return;
		}
	
		// Function to collect all layers including nested groups
		function getAllLayersExceptSelected(doc) {
			var allLayers = [];
			function gather(layers) {
				for (var j = 0; j < layers.length; j++) {
					var layer = layers[j];
					if (layer == doc.activeLayer) {
						continue;
					}
					allLayers.push(layer);
					if (layer.typename === "LayerSet") {
						gather(layer.layers);
					}
				}
			}
			gather(doc.layers);
			return allLayers;
		}
	
		var allLayers = getAllLayersExceptSelected(doc);
		var layers_visibility_arr = [];
	
		// Store layer visibility
		for (var i = 0; i < allLayers.length; i++) {
			layers_visibility_arr[i] = allLayers[i].visible;
		}
		// Hide all layers
		for (var i = 0; i < allLayers.length; i++) {
			allLayers[i].visible = false; // Hide all initially
		}
	
		// Handle regular layer export
		var currentParent = activeLayer.parent;
		while (currentParent && currentParent.typename === "LayerSet") {
			currentParent.visible = true;
			currentParent = currentParent.parent;
		}
	
		// Set full opacity temporarily if not background
		var oldOpacity = activeLayer.opacity;
		if (!activeLayer.isBackgroundLayer) {
			activeLayer.opacity = 100;
		}
	
		// ... [File testing code remains unchanged] ...
	
		// Export document
		var pngFile = new File(rec['path'] + "/" + rec['f_ps2em'] + ".png");
		var opts = new PNGSaveOptions();
		doc.saveAs(pngFile, opts, true);
	
		// Restore original visibility for all layers
		for (var j = 0; j < allLayers.length; j++) {
			allLayers[j].visible = layers_visibility_arr[j];
		}
	
		// Restore original opacity
		if (!activeLayer.isBackgroundLayer) {
			activeLayer.opacity = oldOpacity;
		}
	
		// Save properties of active layer/group
		propFile = new File(rec['path'] + "/" + rec['f_ps2em'] + ".dat");
		propFile.encoding = "BINARY"; 
		propFile.open("w");
		var opacity = activeLayer.opacity/100.0;
		propFile.write(
			"name=" + activeLayer.name + "\n" +
			"opacity = " + opacity + "\n" +
			"visibility = " + activeLayer.visible + "\n" +
			"blendmode = " + String(activeLayer.blendMode).substring(10) + "\n" +
			"data_complete = 1"
		);
		propFile.close();
	
		exportToEmApp_var_out = out;
	};
	//--------------------------------------------------
	// importFromEmApp()
	//--------------------------------------------------
	// shared variables:
	var importFromEmApp_var_out = "";
	var importFromEmApp_var_propArray = new Array();
	// implementation:
	ext.importFromEmApp = function (rec)
	{
		ext.importFromEmApp_impl(rec);
		return JSON.stringify({out:importFromEmApp_var_out});
	}
	ext.importFromEmApp_impl = function (rec)
	{
		//if (ext.data.debug) { alert("importFromEmApp BEGIN()"); }
		var out = "";
		function DBG(msg) {
			if (ext.data.debug) {
				out += 'importFromEmApp(): ' + msg + '\n';
			}
		}
		function DBGA(msg) {
			if (ext.data.debug) {
				if (0) { alert('importFromEmApp(): ' + msg); }
				else   { DBG(msg); }
			}
		}

		DBGA("BEGIN");

		DBG("rec:"+rec);
		DBG("rec.stringify:"+JSON.stringify(rec, null, 2));

		DBGA("1");
		// load properties
		var propArray = new Array(); //properties.split("\n");
		{
			var propFile = File(rec['path'] + "/" + rec['f_em2ps'] + "_processing.dat");
			//propFile.encoding = "UTF8"; 
			propFile.encoding = "BINARY"; 
			//propFile.open("r", "TEXT");
			propFile.open("r");
			var props = propFile.read();
			if (props == "") {
				if (ext.data.debug) {
					alert("Empty prop file");
					//DBGA("Empty prop file");
				}
				cleanup();
				return;
			}
			propFile.close();
			//alert("props:"+props);
			var propLines = props.split("\n");
			for (var i=0; i<propLines.length; i++) {
				var a = propLines[i].split('=');
				if (a.length != 2) continue;
				propArray[a[0].trim()] = a[1].trim(); 
			}
		}

		DBGA("11");
		app.refresh(); // just to be sure that everything is synchronized

		DBGA("12");

		// create a new document if there is none
		if (app.documents.length == 0) {
			DBG("13");
			var width  = propArray['width' ];
			var height = propArray['height'];
			DBG("create a new document size: "+width+"x"+height);
			app.documents.add(width+' px', height+' px', 72, "Escape Motions");
		}

		DBGA("14")
		importFromEmApp_var_out       = out;
		importFromEmApp_var_propArray = propArray;
		app.activeDocument.suspendHistory('Import from an Escape Motions application','ext.importFromEmApp_impl2('+JSON.stringify(rec)+');'); // Everything within the function call is put in a single history item
	}
	ext.importFromEmApp_impl2 = function (rec)
	{
		var out       = importFromEmApp_var_out;
		var propArray = importFromEmApp_var_propArray;
		function DBG(msg) {
			if (ext.data.debug) {
				out += 'importFromEmApp(): ' + msg + '\n';
			}
		}
		function DBGA(msg) {
			if (ext.data.debug) {
				if (0) { alert('importFromEmApp(): ' + msg); }
				else   { DBG(msg); }
			}
		}

		DBG("rec:"+rec);
		DBG("rec.stringify:"+JSON.stringify(rec, null, 2));

		var prev_ruler_units = app.preferences.rulerUnits;
		DBG("before cleanup() units:"+prev_ruler_units);

		function cleanup()
		{
			DBG("cleanup() units:"+prev_ruler_units);
			app.preferences.rulerUnits = prev_ruler_units;
		};
		
		app.preferences.rulerUnits = Units.PIXELS;
		DBGA("2");

		var origDoc = app.activeDocument; 

		DBGA("3");
		// load png to new document
		var pngFile = new File(rec['path'] + "/" + rec['f_em2ps'] + "_processing.png");
		if (!pngFile.exists) {
			DBG("31");
			if (ext.data.debug) { alert("PNG file does not exist!"); }
			cleanup();
			return;
		}
		DBGA("4");
		app.load(pngFile);
		DBGA("5");
		var tmpDoc = app.activeDocument;
		var layer_pos_x  = Number(String(tmpDoc.activeLayer.bounds[0]).split(' ')[0]);
		var layer_pos_y  = Number(String(tmpDoc.activeLayer.bounds[1]).split(' ')[0]);
		var layer_width  = (String(tmpDoc .width ).split(' ')[0]);
		var layer_height = (String(tmpDoc .height).split(' ')[0]);
		var doc_width    = (String(origDoc.width ).split(' ')[0]);
		var doc_height   = (String(origDoc.height).split(' ')[0]);
		DBG("layer pos_x:"+layer_pos_x.toString()+" pos_y:"+layer_pos_y.toString()+" w:"+layer_width+" h:"+layer_height);
		DBG("doc w:"+doc_width+" h:"+doc_height);
		DBG("import properties w:"+propArray['width']+" h:"+propArray['height']);
		DBGA("6");

		DBGA("7");

//		DBG("app.load(): " + pngFile.url + " activeDocument:" + tmpDoc.fullName + " layers:" + tmpDoc.artLayers.length);
//		DBG("app.load(): width:"+tmpDoc.width+" height:"+tmpDoc.height);
//		DBG("app.load(): bounds:" + tmpDoc.activeLayer.bounds);
//		DBG("app.load(): bounds width: " + (tmpDoc.activeLayer.bounds[2].valueOf() - tmpDoc.activeLayer.bounds[0].valueOf()));
//		DBG("app.load(): bounds height:" + (tmpDoc.activeLayer.bounds[3].valueOf() - tmpDoc.activeLayer.bounds[1].valueOf()));
		//var layer_pos_x  = int(String(tmpDoc.activeLayer.bounds[0]).split(' ')[0]);

//		DBG("switch to temp document: " + pngFile.url + " activeDocument:" + app.activeDocument.fullName + " layers:" + app.activeDocument.artLayers.length);

//		DBG("tmpDoc: activeDocument:" + tmpDoc.fullName + " layers:" + tmpDoc.artLayers.length);
		// create a new document if needed
		
		// Cut and close the imported document
		{
			// cut png file as a layer
			DBGA("71");
			DBGA("tmpDoc name:"+tmpDoc.name);
			var tmpLayer = tmpDoc.artLayers[0];
			DBGA("72");
			//tmpDoc.selection.selectAll();
			//DBGA("721");
			tmpLayer.copy();
			DBGA("73");
			// close temp. document
			tmpDoc.close(SaveOptions.DONOTSAVECHANGES);
			DBGA("74");
			DBG("tmpDoc close");
			DBGA("75");
			// switch to origin document
			//tmpDoc = origDoc;
		}
		DBGA("8");
		app.refresh(); // without this it is not properly synchronized in non-tabbed mode
//		origDoc = app.activeDocument;
		
//		DBG("tmpDoc: " + tmpDoc);
//		alert("origDoc: " + JSON.stringify(origDoc));
//		alert("tmpDoc: activeDocument:" + JSON.stringify(origDoc.fullName));
//		DBG("tmpDoc: activeDocument:" + origDoc.fullName);
//		DBG("tmpDoc: layers:" + tmpDoc.artLayers);
//		DBG("tmpDoc: layers length:" + tmpDoc.artLayers.length);
		
		// Paste and translate
		origDoc.paste();
		// Translate
		{
			DBG("png bounds: "+origDoc.activeLayer.bounds);
			var x     = Number(String(origDoc.activeLayer.bounds[0]).split(' ')[0]);
			var y     = Number(String(origDoc.activeLayer.bounds[1]).split(' ')[0]);
			var x_max = Number(String(origDoc.activeLayer.bounds[2]).split(' ')[0]);
			var y_max = Number(String(origDoc.activeLayer.bounds[3]).split(' ')[0]);

			// Move layer's top left corner to the center of the document
			var diff_x = (x_max - x + 1) / 2;
			var diff_y = (y_max - y + 1) / 2;
			DBG("translate x:"+diff_x+" y:"+diff_y);
			origDoc.activeLayer.translate(diff_x, diff_y);
			// Move layer's top left corner to the same position as in the imported png document
			diff_x = - ((layer_width  / 2) - layer_pos_x);
			diff_y = - ((layer_height / 2) - layer_pos_y);
			DBG("translate x:"+diff_x+" y:"+diff_y);
			origDoc.activeLayer.translate(diff_x, diff_y);

			DBG("png bounds x:"+x+" y:"+y);
			DBG("png bounds after translation: "+origDoc.activeLayer.bounds);
		}
		
		DBGA("9");
		//DBG("origDoc paste: activeDocument:" + origDoc.fullName + " layers:" + origDoc.artLayers.length);
		
		// set properties
		// set layer name
		origDoc.activeLayer.name = propArray['name'];
		// set opacity
		origDoc.activeLayer.opacity = Number(propArray['opacity']) * 100.0;
		// set visibility
		origDoc.activeLayer.visible = propArray['visibility'] != "0";
		// set blendmode	
		DBGA("91");
		var blendmode = propArray['blendmode']; 
		if (blendmode == "DARKEN") {
			origDoc.activeLayer.blendMode = BlendMode.DARKEN;
		} else if (blendmode == "LIGHTEN") {
			origDoc.activeLayer.blendMode = BlendMode.LIGHTEN;
		//} else if (blendmode == "ADDITIVE") {
		//	origDoc.activeLayer.blendMode = BlendMode.ADD;
		} else if (blendmode == "DIFFERENCE") {
			origDoc.activeLayer.blendMode = BlendMode.DIFFERENCE;
		} else if (blendmode == "HARD_LIGHT") {
			origDoc.activeLayer.blendMode = BlendMode.HARDLIGHT;
		} else if (blendmode == "MULTIPLY") {
			origDoc.activeLayer.blendMode = BlendMode.MULTIPLY;
		} else if (blendmode == "OVERLAY") {
			origDoc.activeLayer.blendMode = BlendMode.OVERLAY;
		} else if (blendmode == "SCREEN") {
			origDoc.activeLayer.blendMode = BlendMode.SCREEN;
		} else if (blendmode == "SUBTRACTIVE") {
			origDoc.activeLayer.blendMode = BlendMode.SUBTRACT;
		} else {
			origDoc.activeLayer.blendMode = BlendMode.NORMAL;
		}
		DBG("blendmode png:"+blendmode+" layer:"+origDoc.activeLayer.blendMode);
		DBGA("10");
		
		DBG("origDoc set properties done");

		// Mismatched dimensions message
		if (layer_width  != doc_width ||
			layer_height != doc_height)
		{
			msgFile = new File(rec['path'] + "/" + rec['f_ps2em'] + ".msg");
			msgFile.encoding = "BINARY";
			msgFile.open("w");
			msgFile.write("PHOTOSHOP_PLUGIN_DIMENSIONS_MISMATCH");
			msgFile.close();
		}

		cleanup();

		DBG("END");
		//return '{"out":"'+out+'","i":'+i+',"msg":"'+msg+'"}';
		var msg = "Ok";
		//return JSON.stringify({out:"ok"});
		//return JSON.stringify({out:out, msg:msg});
		importFromEmApp_var_out = out;
	};


    //Evaluate a file and catch the exception.
    ext.evalFile = function (path) {
        try {
            $.evalFile(path);
        } catch (e) {
            alert("Exception:" + e);
        }
    };
    
    // Evaluate all the files in the given folder 
    ext.evalFiles = function (jsxFolderPath) {
        var folder = new Folder(jsxFolderPath);
        if (folder.exists) {
            var jsxFiles = folder.getFiles("*.jsx");
            var i;
            for (i = 0; i < jsxFiles.length; i++) {
                var jsxFile = jsxFiles[i];
                $._ext.evalFile(jsxFile);
            }
        }
    };

	ext.init = function (dataIn)
	{
		//alert("hostscript init("+dataIn.extPath+")");
		ext.evalFile(dataIn.extPath + "/jsx/json2.js");
		//alert("data:"+JSON.stringify(dataIn));
		ext.data = dataIn;
		return JSON.stringify(dataIn);
	};

    return ext;
}());
